/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.crypt;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Base64;

import com.graveng.rci.util.Logger;

/**
 * パケットを暗号化するかのフィルター
 */
public class CryptFilter {

	/**
	 * 公開鍵、秘密鍵のペア
	 */
	public RSAKeyPairCrypt pairCrypt = new RSAKeyPairCrypt();

	// packetはUTF-8だが、普通のStringでgetBytesするとWindows標準環境ではMS932になるので、
	// それを電文に載せるとUTF-8と混在しておかしくなるようだ。
	// -Dfile.encoding=UTF8 とするしかないようだ。

	/**
	 * 複数回の暗号化／復号化を繰り返す
	 */
	protected ArrayList<AesCrypt> aesCrypts = new ArrayList<AesCrypt>();

	/**
	 * 暗号化状態
	 */
	protected boolean isAesCryptMode = false;

	/**
	 * @return isAesCryptMode
	 */
	public boolean isAesCryptMode() {
		return isAesCryptMode;
	}

	/**
	 * @param isAesCryptMode セットする isAesCryptMode
	 */
	public void setAesCryptMode(boolean isAesCryptMode) {
		this.isAesCryptMode = isAesCryptMode;
	}

	/**
	 * AesCryptを追加する。（複数回の暗号化／復号化対応）
	 * @param aescrypt AesCrypt
	 */
	public void addAesCrypt(AesCrypt aescrypt) {
		aesCrypts.add(aescrypt);
	}

	/**
	 * RSAの公開鍵を取得する。
	 * @return 公開鍵
	 * @throws Exception 何らかの例外
	 */
	public String getPublicKey() throws Exception {
		if(pairCrypt.getPublicKey() == null) {
			pairCrypt.createKeyPairAsBase64();
		}
		return pairCrypt.getPublicKey();
	}

	/**
	 * 状態に応じて、暗号化する。
	 * @param data 平文
	 * @return 暗号文か平文
	 * @throws IOException IO例外
	 */
	public String encrypt(String data) throws IOException {
		Logger.debug(data);
		String result = null;
		if(isAesCryptMode) {
			byte[] bytes = data.getBytes();
			try {
				for(int i = 0; i < aesCrypts.size(); i++) {
					// 複数回の暗号化はバイナリからバイナリへの変更
					AesCrypt crypter = aesCrypts.get(i);
					bytes = crypter.encrypt(bytes);
				}
			} catch (Exception e) {
				throw new IOException("failed AES crypt", e);
			}
			// Base64でテキスト化する。
			result = new String(Base64.getEncoder().encode(bytes));
		}
		else {
			result = data;
		}

		Logger.verbose(result);
		return result;
	}

	/**
	 * 状態に応じて、復号化する。
	 * @param data 平文か暗号文
	 * @return 平文
	 * @throws IOException IO例外
	 */
	public String decrypt(String data) throws IOException {
		Logger.verbose(data);
		String result = null;
		if(isAesCryptMode) {
			// Base64のテキストをバイナリに戻す
			byte[] bytes = Base64.getDecoder().decode(data.getBytes());
			try {
				for(int i = aesCrypts.size()-1; i >= 0 ; i--) {
					// 複数回の復号化はバイナリからバイナリへの変更で解読されても特徴がないので解読できたと認識不能
					AesCrypt crypter = aesCrypts.get(i);
					bytes = crypter.decrypt(bytes);
				}
			} catch (Exception e) {
				throw new IOException("failed AES decrypt", e);
			}
			result = new String(bytes);
		}
		else {
			result = data;
		}
		Logger.debug(result);
		return result;
	}

}
