/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.crypt;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SecureRandom;
import java.security.Security;
import java.util.Base64;
import java.util.Date;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import com.graveng.rci.util.Logger;


// org.bouncycastle.jce.provider.BouncyCastleProvider用にbcprov-jdk15on-159.jarを使っている
/**
 * RSA暗号の公開鍵と秘密鍵の生成、暗号化、復号化を行う。
 * 暗号化ライブラリはorg.bouncycastle.jce.provider.BouncyCastleProviderを使用している。
 */
public class RSAKeyPairCrypt {
	static {
        Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
	}

	/**
	 * キー長、デフォルトは2048。<br>
	 * 1024や512も指定できるが、暗号化速度は速くなるが解読される脆弱性があるらしい。<br>
	 * 接続時にテンポラリのキーペアを生成し、AES暗号の共通鍵と初期ベクターを暗号化する用途で使うだけなら
	 * 毎回変わるので解読するのも無意味であるから、接続速度を上げるためには、512や1024にするのも有効性が
	 * あるかも知れないが、処理速度的に2秒が1秒に変わる程度なので、やはり2048のまま変えない方がよい。
	 */
	protected static final int KeyLength = 2048;

	/**
	 * base64にエンコードした公開鍵を格納
	 */
	protected String publicKey = null;

	/**
	 * base64にエンコードした秘密鍵を格納
	 */
	protected String secretKey = null;

	/**
	 * デフォルトコンストラクタ
	 */
	public RSAKeyPairCrypt() {

	}

	/**
     * base64にエンコードした公開鍵を取得する。<br>
     * <br>
     * 初期状態はnullでcreateKeyPairAsBase64()にて生成後に取得する。
	 * @return 公開鍵
	 */
	public String getPublicKey() {
		return publicKey;
	}

	/**
	 * base64にエンコードした公開鍵を設定する。
	 * @param publicKey 公開鍵
	 */
	public void setPublicKey(String publicKey) {
		this.publicKey = publicKey;
	}

	/**
     * base64にエンコードした秘密鍵を取得する。
     * <br>
     * 初期状態はnullでcreateKeyPairAsBase64()にて生成後に取得する。
	 * @return 秘密鍵
	 */
	public String getSecretKey() {
		return secretKey;
	}

	/**
	 * base64にエンコードした秘密鍵を設定する。
	 * @param secretKey 秘密鍵
	 */
	public void setSecretKey(String secretKey) {
		this.secretKey = secretKey;
	}

	/**
     * RSAキーペアを生成する。
	 * @throws Exception 何らかの例外
     */
    public void createKeyPairAsBase64() throws Exception{
        KeyPairGenerator generator;
        try {
            generator = KeyPairGenerator.getInstance("RSA", "BC");
        } catch (NoSuchAlgorithmException | NoSuchProviderException ex) {
            Logger.error("createKeyPairAsBase64 failed", ex);
            return;
        }
        SecureRandom random = new SecureRandom();
        generator.initialize(KeyLength, random);
        KeyPair pair = generator.generateKeyPair();

        // KeyPairをBase64エンコードして返却
        publicKey = encodeObjectAsBase64(pair.getPublic());
        secretKey = encodeObjectAsBase64(pair.getPrivate());
    }

    /**
     * 公開鍵で暗号化する。
     * @param data 暗号化前
     * @return 暗号化後
	 * @throws Exception 何らかの例外
     */
    public byte[] encryptByPublicKey(byte[] data) throws Exception {
    	return encryptByBase64Key(data, publicKey);
    }

    /**
     * 秘密鍵で暗号化する。
     * @param data 暗号化前
     * @return 暗号化後
	 * @throws Exception 何らかの例外
     */
    public byte[] encryptBySecretKey(byte[] data) throws Exception{
    	return encryptByBase64Key(data, secretKey);
    }

    /**
     * 公開鍵で復号化する。
     * @param data 復号化前
     * @return 復号化後
	 * @throws Exception 何らかの例外
     */
    public byte[] decryptByPublicKey(byte[] data) throws Exception{
    	return decryptByBase64Key(data, publicKey);
    }

	/**
	 * 秘密鍵で復号化する。
	 *
	 * @param data
	 *            復号化前
	 * @return 復号化後
	 * @throws Exception
	 *             何らかの例外
	 */
	public byte[] decryptBySecretKey(byte[] data) throws Exception {
		return decryptByBase64Key(data, secretKey);
	}

    /**
     * キーを与え暗号化を行う。
     * @param data 暗号化前
     * @param base64Key base64にエンコードされた公開鍵か秘密鍵
     * @return 暗号化後
	 * @throws Exception 何らかの例外
     */
    public byte[] encryptByBase64Key(byte[] data, String base64Key) throws Exception {
        try {
            Key myKey = (Key) decodeObjectFromBase64(base64Key);
            // 117バイトまでしか暗号化できないらしい。
            // https://stackoverflow.com/questions/2580473/rsa-example-that-do-not-use-nopadding
            // Cipher.getInstance("RSA/ECB/PKCS1PADDING");

            Cipher cipher = Cipher.getInstance("RSA/None/NoPadding", "BC");
//            Cipher cipher = Cipher.getInstance("RSA/ECB/PKCS1PADDING", "BC");
            cipher.init(Cipher.ENCRYPT_MODE, myKey, new SecureRandom());
            return cipher.doFinal(data);
        } catch (IOException | NoSuchAlgorithmException | NoSuchProviderException | NoSuchPaddingException | IllegalBlockSizeException | BadPaddingException | InvalidKeyException  ex) {
//            Logger.error("catch exception", ex);
            throw ex;
        }
    }

    /**
     * キーを与え復号化を行う。
     * @param data 復号化前
     * @param base64Key base64にエンコードされた公開鍵か秘密鍵
     * @return 復号化後
	 * @throws Exception 何らかの例外
     */
    public byte[] decryptByBase64Key(byte[] data, String base64Key) throws Exception {
        try {
            Key myKey = (Key) decodeObjectFromBase64(base64Key);
            Cipher cipher = Cipher.getInstance("RSA/None/NoPadding", "BC");
//            Cipher cipher = Cipher.getInstance("RSA/ECB/PKCS1PADDING", "BC");
            cipher.init(Cipher.DECRYPT_MODE, myKey, new SecureRandom());
            return cipher.doFinal(data);
        } catch (IOException | NoSuchAlgorithmException | NoSuchProviderException | NoSuchPaddingException | IllegalBlockSizeException | BadPaddingException | InvalidKeyException  ex) {
//            Logger.error("catch exception", ex);
            throw ex;
        }
    }

	/**
	 * ObjectをBase64にエンコードする。
	 *
	 * @param o
	 *            Object
	 * @return エンコード結果
	 */
	protected String encodeObjectAsBase64(Object o) {
		try (ByteArrayOutputStream byteos = new ByteArrayOutputStream();
				GZIPOutputStream gos = new GZIPOutputStream(byteos);) {
			try (ObjectOutputStream objos = new ObjectOutputStream(gos)) {
				objos.writeObject(o);
			}
			byte[] retObject = byteos.toByteArray();
			return Base64.getEncoder().encodeToString(retObject);
		} catch (IOException ex) {
			Logger.warn("catch exception", ex);
			return null;
		}
	}

	/**
	 * Base64エンコードされたObjectをObjectに変換する。
	 *
	 * @param s
	 *            Base64エンコードされたObject
	 * @return Object
	 * @throws IOException
	 *             何らかの例外
	 */
	protected Object decodeObjectFromBase64(String s) throws IOException {
		byte[] bytes = Base64.getDecoder().decode(s);
		try (GZIPInputStream gis = new GZIPInputStream(new ByteArrayInputStream(bytes))) {
			return new ObjectInputStream(gis).readObject();
		} catch (ClassNotFoundException ex) {
			Logger.warn("catch exception", ex);
			return null;
		}
	}

    /**
     * RSA暗号の公開鍵と秘密鍵を生成し、暗号化、復号化可能か検証する。
     * @param argv なし
     */
    public static void main(String argv[]){

		try {
	        RSAKeyPairCrypt sample = new RSAKeyPairCrypt();
	        System.out.println("【KeyPairの生成】");

	        // org.bouncycastle.jce.provider.BouncyCastleProviderは何か遅い。２秒ぐらいかかる。
	        System.out.println(new Date());
	        System.out.println("---生成中...---");
	        sample.createKeyPairAsBase64();
	        System.out.println(new Date());

	        System.out.println("----------------------------------------");
	        System.out.println("公開鍵 ＞ " + sample.getPublicKey());
	        System.out.println("秘密鍵 ＞ " + sample.getSecretKey());
	        System.out.println("----------------------------------------");

	        System.out.println("【生成したKeyPairの検証】");
	        final String message = "＜RSAKeyPair暗号化の動作確認＞";
	        byte[] encryptBytes;
				encryptBytes = sample.encryptByBase64Key(message.getBytes(), sample.getSecretKey());
	        System.out.println("【平文】");
	        System.out.println(message);
	        System.out.println("【秘密鍵で暗号化】");
	        System.out.println(sample.encodeObjectAsBase64(encryptBytes));

	        byte[] decryptBytes = sample.decryptByBase64Key(encryptBytes, sample.getPublicKey());
	        System.out.println("【公開鍵で復号化】");
	        System.out.println(new String(decryptBytes));

	        System.out.println("【公開鍵で暗号化】");
	        byte[] encryptBytes2 = sample.encryptByPublicKey(message.getBytes());
	        System.out.println(sample.encodeObjectAsBase64(encryptBytes2));

	        System.out.println("【秘密鍵で復号化】");
	        byte[] decryptBytes2 = sample.decryptBySecretKey(encryptBytes2);
	        System.out.println(new String(decryptBytes2));
		} catch (Exception e) {
			e.printStackTrace();
		}

    }

}