/*
 * Copyright (C) 2018 Gravity Engineering Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.graveng.rci.server;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.management.ManagementFactory;
import java.lang.management.RuntimeMXBean;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Base64;
import java.util.EventListener;
import java.util.HashMap;

import org.apache.commons.io.FileUtils;

import com.graveng.rci.NullObject;
import com.graveng.rci.Packet;
import com.graveng.rci.RciException;
import com.graveng.rci.crypt.AesCrypt;
import com.graveng.rci.crypt.CryptFilter;
import com.graveng.rci.crypt.HashCrypt;
import com.graveng.rci.server.config.GroupConfig;
import com.graveng.rci.server.config.IpAddressConfig;
import com.graveng.rci.server.config.ServerConfig;
import com.graveng.rci.server.config.UserConfig;
import com.graveng.rci.util.ClassField;
import com.graveng.rci.util.Logger;
import com.graveng.rci.util.PlatformUtils;
import com.graveng.rci.util.StringUtils;

/**
 * RCIサーバメイン。クライアントからの接続を受け付け、クラスを起動する。
 */
public class RciServer {
//	static boolean isWrappedException = false;
	/**
	 * サーバで発生した例外をそのままクライアントに送るならfalse, 例外クラス名とメッセージを送るならtrue
	 */
	protected boolean isWrappedException = Cm.gi().getConfig().getServerConfig().isWrappedException();

	/**
	 * 設定ファイルがあるディレクトリ
	 */
	protected String baseDirectory = Cm.gi().getBaseDirectory();

	/**
	 * サーバソケットで待ち受けるスレッド
	 */
	protected Accepter accepter = null;

	/**
	 * 起動時パラメータを与え、RciServerを起動する。起動方法はServerConfig.xmlで設定する。
	 * @param args mainのパラメータ<br>
	 * [-p ポート番号] [-s 起動時パスワード]
	 * @throws Exception 何らかの例外
	 */
	public RciServer(String[] args)  throws Exception {
        // optionの取得を行う
        HashMap<String, String> options = StringUtils.getOptions(args);

        ServerConfig sc = Cm.gi().getConfig().getServerConfig();
        // ポート番号の取得
		int port;
        String portNumber = options.get("-p");
        if(portNumber != null) {
			port = Integer.parseInt(args[0]);
        }
        else {
			port = sc.getPort(); // デフォルトポート番号
        }

        if(sc.isServerCertification()) {
			// プロセス起動認証用にパスワードの照合を行う
	        String serverCertification = options.get("-s");
	        if(serverCertification == null) {
	        	throw new IllegalArgumentException("need password for certification");
	        }
	        else {
	        	HashCrypt crypt = new HashCrypt();
	        	boolean isCertificate = crypt.certificate(sc.getHashAlgorithmName(), sc.getPasswordAuthHashRoundCount(), serverCertification, sc.getPasswordRegistHashRoundCount(), sc.getPassword());
	        	if(!isCertificate) {
	        		throw new IllegalArgumentException("certification failure");
	        	}
	        }
        }
		savePid();
		bind(port);
		initializeClass();
	}

	/**
	 * ユーザ定義の起動時クラスを起動する。
	 * @throws Exception 何らかの例外
	 */
	protected void initializeClass() throws Exception {
		String initClass = Cm.gi().getConfig().getServerConfig().getInitializeClass();
		if(!StringUtils.isEmpty(initClass)) {
			Class<?> c = Class.forName(initClass);
			Object object = c.newInstance();
			if(object instanceof RciServerInitialize) {
				RciServerInitialize init = (RciServerInitialize)object;
				init.initialize(this);
			}
		}

	}

	/**
	 * サーバソケットで待ち受けるAccepterを起動する。
	 * @param port ポート番号
	 * @throws IOException IO例外
	 */
	protected void bind(int port) throws IOException {
		accepter = new Accepter(port);
		Thread thread = new Thread(accepter);
		thread.start();
	}

	/**
	 * サーバソケットをクローズする。
	 * @throws IOException IO例外
	 */
	public void close() throws IOException {
		if(accepter != null) { accepter.disconnect(); }
	}

	/**
	 * 終了バッチ、終了シェルをファイルに保存する。
	 * @throws IOException IO例外
	 */
	protected void makeKillme() throws IOException {
		if(PlatformUtils.isWindows()) {
			long pid = getPid();
			File file = new File(baseDirectory + "/killme.bat");

			// windows special (unix なら killコマンド) /Fの強制終了をつけないと終わらないようだ
			String cmd = "taskkill /F /pid "+ pid;
			FileUtils.write(file, cmd);
		}
		else {
			long pid = getPid();
			File file = new File(baseDirectory + "/killme.sh");

			// windows special (unix なら killコマンド)
			String cmd = "kill "+ pid;
			FileUtils.write(file, cmd);
		}
	}

	/**
	 * 自プロセスのプロセスIDをファイルに保存する。
	 * @throws IOException IO例外
	 */
	protected void savePid() throws IOException {
		long pid = getPid();
		File file = new File(baseDirectory + "/pid.txt");

		String cmd = ""+ pid;
		FileUtils.write(file, cmd);

		makeKillme();
	}

	/**
	 * 自プロセスのプロセスIDを取得する。
	 * @return プロセスID
	 */
	protected long getPid() {
		RuntimeMXBean bean = ManagementFactory.getRuntimeMXBean();
	    String vmName = bean.getName();
	    long pid = Long.valueOf(vmName.split("@")[0]);
	    return pid;
	}

	/**
	 * サーバソケットで待ち受けるAccepterを取得する。
	 * @return accepter
	 */
	public Accepter getAccepter() {
		return accepter;
	}

	/**
	 * 起動時パラメータを与え、RciServerを起動する。起動方法はServerConfig.xmlで設定する。
	 * @param args mainのパラメータ<br>
	 * [-p ポート番号] [-s 起動時パスワード]
	 */
	public static void main(String[] args) {
       try {
    	   RciServer server = new RciServer(args);

	        // shutdownフックの登録
	        Runtime.getRuntime().addShutdownHook(new Thread(
	                () -> {
	                	// こんなの登録しても呼ばれない場合がある。
	                	// unixではSIGTERMでは呼ばれて、SIGKILLなら呼ばれない。

	                	// ここが正常終了
	                    System.out.println("shutdown ...");
	                    try {
							server.close();
						} catch (IOException e) {
							e.printStackTrace();
						}
	                }
	                ));

        } catch (Exception e) {
			e.printStackTrace();
		}

		// 止めるためにはシグナル送信
	}

	/**
	 * サーバソケットで待ち受け、コネクション毎のスレッドを生成する
	 */
	public class Accepter implements Runnable {
		/**
		 * サーバ待ち受けるポート番号
		 */
		protected int port;

		/**
		 * サーバソケット
		 */
		protected ServerSocket ss;

		/**
		 * 接続されているクライアント<br>
		 * key:クライアント向けのSocket, value:個々のクライアントと送受信するInvoker
		 */
		protected HashMap<Socket, Invoker> invokers = new HashMap<Socket, Invoker>();

		/**
		 * ログインに失敗したユーザの保持。
		 * key:ユーザ名, value:LoginFailureUser
		 */
		protected HashMap<String, LoginFailureUser> failLogin = new HashMap<String, LoginFailureUser>();

		public Accepter() {

		}

		public Accepter(int port) throws IOException {
			bind(port);
		}

		public void bind(int port) throws IOException {
			this.port = port;
			ss = new ServerSocket(port);
		}
		public int getPort() {
			return port;
		}

		/**
		 * クライアントとの接続が切れた際に接続しているクライアントの情報を削除する。
		 * @param s Socket
		 */
		protected void closeInvoker(Socket s) {
			invokers.remove(s);
		}

		/**
		 * クライアントとの接続情報を登録する。
		 * @param s クライアントソケット
		 * @param invoker Invoker
		 */
		protected void putInvoker(Socket s, Invoker invoker) {
			invokers.put(s, invoker);
		}

		/**
		 * サーバソケットで待ち受け、クライアント毎の通信を行うInvokerスレッドを起動する。
		 */
		public void run() {
			try {
		        while(true) {
		            Socket s = ss.accept(); //クライアントからの通信開始要求が来るまで待機

		            Logger.info("accept from " + s.getRemoteSocketAddress());
		            Invoker invoker = new Invoker(s, this);

		            if(!isAccessibleIpAddress(s, invoker)) {
		            	invoker.close();
		            	continue;
		            }
		            else {
			            Thread th = new Thread(invoker);
			            th.start();
		            }
		        }
			} catch (Exception e) {
				Logger.info("accepter close", e);
			}
		}

		/**
		 * クライアントのIPアドレスが接続可能か判定する。
		 * @param s クライアントソケット
		 * @param invoker Invoker
		 * @return trueの場合接続可能
		 * @throws IOException IO例外
		 */
		protected boolean isAccessibleIpAddress(Socket s, Invoker invoker) throws IOException {
			boolean result;
			// IpAddress定義を取得

			if(Cm.gi().getConfig().getIpAddressDenyFilter().isEmpty()) {
				// 不許可定義がない
				if(Cm.gi().getConfig().getIpAddressAllowFilter().isEmpty()) {
					// 不許可(0)、許可(0)→全許可
					result = true;
				}
				else {
					IpAddressConfig isAccess = getIpAddressConfig(Cm.gi().getConfig().getIpAddressAllowFilter(), s.getInetAddress().getHostAddress());
					if(isAccess == null) {
						// デフォルト定義を取得
			        	isAccess = Cm.gi().getConfig().getIpAddressAllowFilter().get("");
			        	if(isAccess == null) {
			        		// 不許可(0)、許可(非0、ハズレ)→不許可
			                Logger.warn("unable connect by ipAddressConfig from:" + s.getInetAddress());
			                result =  false;
			        	}
			        	else {
			        		// 不許可(0)、許可(非0、デフォルト)→許可
			            	invoker.encryptPacket = isAccess.isEncryptPacket();
			            	invoker.userCertification = isAccess.isUserCertification();
			            	result = true;
			        	}
			        }
					else {
						// 不許可(0)、許可(非0、ヒット)→許可
			        	invoker.encryptPacket = isAccess.isEncryptPacket();
			        	invoker.userCertification = isAccess.isUserCertification();
			        	result = true;
					}
				}
			}
			else {
				IpAddressConfig isAccess = getIpAddressConfig(Cm.gi().getConfig().getIpAddressDenyFilter(), s.getInetAddress().getHostAddress());
				if(isAccess == null) {
					isAccess = getIpAddressConfig(Cm.gi().getConfig().getIpAddressAllowFilter(), s.getInetAddress().getHostAddress());
					if(isAccess == null) {
						// デフォルト定義を取得
			        	isAccess = Cm.gi().getConfig().getIpAddressAllowFilter().get("");
			        	if(isAccess == null) {
			        		// 不許可(ハズレ)、許可(非0、ハズレ)→不許可
			                Logger.warn("unable connect by ipAddressConfig from:" + s.getInetAddress());
			                result =  false;
			        	}
			        	else {
			        		// 不許可(ハズレ)、許可(非0、デフォルト)→許可
			            	invoker.encryptPacket = isAccess.isEncryptPacket();
			            	invoker.userCertification = isAccess.isUserCertification();
			            	result = true;
			        	}
			        }
					else {
						// 不許可(ハズレ)、許可(非0、ヒット)→許可
			        	invoker.encryptPacket = isAccess.isEncryptPacket();
			        	invoker.userCertification = isAccess.isUserCertification();
			        	result = true;
					}
				}
				else {
					// 不許可(非0、ヒット)→不許可（許可と不許可に同じkeyがあるなら、不許可を優先
	                Logger.warn("unable connect by ipAddressConfig from:" + s.getInetAddress());
	                result =  false;
				}
			}
			return result;
		}

		/**
		 * クライアントのIPアドレスに対応するIpAddressConfigを取得する。
		 * @param map 設定ファイルからロードされたマップ(key:IPアドレス、value:IpAddressConfig)
		 * @param ipAddressClient クライアントのIPアドレス
		 * @return IpAddressConfig
		 */
		protected IpAddressConfig getIpAddressConfig(HashMap<String, IpAddressConfig> map, String ipAddressClient) {
			String ipAddress;
//			String separator = "\\.";
			String separator = ".";
			// IPv6の:は.に変換する(ファイル名にできないため)
			ipAddress = ipAddressClient.replaceAll(":", ",");
			if(!ipAddressClient.equals(ipAddress)) {
				// IPv6の場合
				separator = ",";
			}

			IpAddressConfig result;
			result = map.get(ipAddress);
			if(result != null) {
				return result;
			}
			String subnet = ipAddress;
			while(!"".equals(subnet)) {
//				subnet = getDotParent(subnet);
				subnet = StringUtils.getSeparatorParent(subnet, separator);
				result = map.get(subnet);
				if(result != null) {
					return result;
				}
			}
			return result;
		}

		/**
		 * サーバソケットをクローズする。
		 * @throws IOException IO例外
		 */
		public void disconnect() throws IOException {
			ss.close();
		}
	}

	/**
	 * １つのコネクションの送受信を行う
	 */
	public class Invoker implements Runnable{
		protected Accepter accepter = null;
		protected Socket socket;
		protected DataInputStream dis = null;
		protected DataOutputStream dos = null;

		// key:インスタンス番号, value:インスタンス
		protected HashMap<Integer, Object> instanceMap = new HashMap<Integer, Object>();

		// key:リスナのハッシュコード, value:クライアントのインスタンス番号
		protected HashMap<Integer, Integer> listenerMap = new HashMap<Integer, Integer>();

		// new 用のインスタンス番号(プラス値)
		protected int iNumber = 0;

		// old（シングルトン)用のインスタンス番号(マイナス値)
		protected int singleNumber = 0;

		/**
		 * 暗号化フィルター
		 */
		protected CryptFilter cryptFilter = new CryptFilter();

		/**
		 * 接続されたユーザ名
		 */
		protected String loginUsername = "anonymous";

		/**
		 * クライアントのIPアドレス
		 */
		protected String remoteAddress = "";

		/**
		 * ユーザ設定情報
		 */
		protected UserConfig userConfig = null;

		/**
		 * 回線を暗号化する場合はtrue
		 */
		protected boolean encryptPacket;

		/**
		 * 回線でユーザ認証を行う場合はtrue
		 */
		protected boolean userCertification;

		/**
		 * １つのコネクションの送受信スレッドの生成
		 * @param socket １つのクライアントとの回線のソケット
		 */
		public Invoker(Socket socket) {
			this.socket = socket;
			init();
		}

		/**
		 * １つのコネクションの送受信スレッドの生成
		 * @param socket １つのクライアントとの回線のソケット
		 * @param accepter クライアントの接続を待ち受ける
		 */
		public Invoker(Socket socket, Accepter accepter) {
			this.socket = socket;
			this.accepter = accepter;
			accepter.putInvoker(socket, this);
			init();
		}

		/**
		 * クライアントとの回線断
		 */
		public void close() {
			try {
				delListener();
	            if(dos != null) dos.close();
	            if(dis != null) dis.close();
	            if(accepter != null) accepter.closeInvoker(socket);
	            if(socket != null) socket.close();
			}
			catch(Exception ie) {
				Logger.warn("close error", ie);
			}
		}

		/**
		 * ソケット入出力の初期化
		 */
		protected void init() {
			try {
				remoteAddress = socket.getInetAddress().getHostAddress();
		        InputStream is = socket.getInputStream();
		        dis = new DataInputStream(is);

		        OutputStream os = socket.getOutputStream();
		        dos = new DataOutputStream(os);
			}
			catch(Exception e) {
				Logger.error("initialize error", e);
			}
		}

		/**
		 * ソケット受信用のスレッド
		 */
		public void run() {
			try {
	    		while(true) {
		            // 入力処理
	    			String req = recv();
		            recieved(req);
	    		}
			}
			catch(Exception e) {
				Logger.info("communicate disconnect user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]", e);
			}
			finally {
				close();
			}
		}

		/**
		 * パケット受信後の処理振り分けを行う
		 * @param text 電文データ
		 * @throws Exception 何らかの例外
		 */
		protected void recieved(String text) throws Exception {
			Packet req = null;
			try {
				req = (Packet)Packet.getDecodeXML(text);
				req.deSerializeParam();
			}
			catch(Exception e) {
				String msg = "illegal packet received user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]";
				Logger.warn(msg, e);
				throw new IllegalArgumentException(msg, e);
			}

			// 鍵交換かユーザ認証かパスワード変更を行うか
			// static メソッドを実行するか
			// インスタンスを生成するか
			// 生成済のインスタンスのメソッドを実行するか破棄するか

			try {
				if(req.getInstanceNumber() == 0) {
					if(req.getParams().length >= 1 && req.getParams()[0].equals("new")) {
						// インスタンスを生成
						checkState();
						invokeNew(req);
					}
					else if(req.getParams().length >= 1 && req.getParams()[0].equals("single")) {
						// シングルトン
						checkState();
						invokeSingle(req);
					}
					else if(req.getParams().length >= 1 && req.getParams()[0].equals("static")) {
						// static メソッドを実行
						checkState();
						invokeStatic(req);
					}
					else if(req.getParams().length >= 1 && req.getParams()[0].equals("ExchangeKey")) {
						if(!encryptPacket) {
							throw new IllegalStateException("encrypt not necessary");
						}

						// 公開鍵受信、公開鍵で暗号化して共通鍵送信
						invokeExchangeKey(req);
						String sendData = Packet.getEncodeXML(req);
						send(sendData);
						cryptFilter.setAesCryptMode(true);
						return;
					}
					else if(req.getParams().length >= 1 && req.getParams()[0].equals("CertificateUser")) {
						if(!userCertification) {
							throw new IllegalStateException("user certification not necessary");
						}

						// ユーザ認証
						invokeCertificateUser(req);
						String sendData = Packet.getEncodeXML(req);
						send(sendData);
						return;
					}
					else if(req.getParams().length >= 1 && req.getParams()[0].equals("UpdatePassword")) {
						if(!userCertification || "anonymous".equals(loginUsername)) {
							throw new IllegalStateException("user certification necessary");
						}

						// パスワード変更
						invokeUpdatePassword(req);
						String sendData = Packet.getEncodeXML(req);
						send(sendData);
						return;
					}
					else {
						throw new IllegalArgumentException("unknown request user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
					}
				}
				else {
					// インスタンスクラスの実行（シングルトンもここ）
					checkState();
					invokeInstance(req);
				}
			}
			catch(Exception e) {
				Logger.warn("invoke failed:", e);
				if(isWrappedException) {
					RciException ex = new RciException(e.getClass().getName() + ":" + e.getMessage());
					req.setException(ex);
				}
				else {
					req.setException(e);
				}
			}

			req.enSerializeParam();
			avoidXMLEncode(req);

			String sendData = Packet.getEncodeXML(req);
			send(sendData);
		}

		/**
		 * 実行可能か判定し不可の場合は例外を投げる
		 */
		protected void checkState() {
			if(userCertification && "anonymous".equals(loginUsername)) {
				throw new IllegalStateException("user certification necessary");
			}
			if(encryptPacket && !cryptFilter.isAesCryptMode()) {
				throw new IllegalStateException("encrypt necessary");
			}
		}

		/**
		 * ソケットから復号化して受信
		 * @return 受信したテキストを復号化したもの
		 * @throws IOException IO例外
		 */
		protected String recv() throws IOException {
			// Stringの場合、パケット全体を一括して受信できるのでそうしている。
			// 中身を覗くと使う気しないが、とりあえず、気にしない。
			String data = dis.readUTF();
	        return  cryptFilter.decrypt(data);
		}

		/**
		 * ソケットへ暗号化して送信
		 * @param text 平文
		 * @throws IOException IO例外
		 */
		protected void send(String text) throws IOException {
			String data = cryptFilter.encrypt(text);
	        dos.writeUTF(data);
		}

		/**
		 * インスタンスクラスのメソッドを実行する。
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
		// methodName param1 param2
//		protected void invokeInstance(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		protected void invokeInstance(Packet req) throws Exception {
			Object[] params;

			params = req.getParams();
			if(params.length == 0) {
				// インスタンスの破棄
				instanceMap.remove(req.getInstanceNumber());

				// リスナー登録されていれば破棄
				removeListener(req);
				req.setInstanceNumber(0);
				return;
			}

			Object result;
			String methodName = params[0].toString();

			Object object = instanceMap.get(req.getInstanceNumber());
			if(object == null) {
				throw new ClassNotFoundException("not found instancee number["+req.getInstanceNumber()+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}
			Class<? extends Object> c = object.getClass();

			String className = c.getName();
			String packageName = ClassField.getPackageName(className);
			if(!isAccessiblePackage(packageName)) {
//				Logger.warn("not allow package["+packageName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
				throw new IllegalAccessException("not allow package["+packageName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}
			if(!isAccessibleClass(className, methodName)) {
//				Logger.warn("not allow className["+className+"] and methodName["+methodName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
				throw new IllegalAccessException("not allow className["+className+"] and methodName["+methodName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			int offset = 1;
			Object[] args = new Object[params.length -offset];
			Class<?>[] paramType = new Class[params.length -offset];
			if(req.getOptionMap().containsKey(Packet.OPTION_PARAMTYPES)) {
				paramType = (Class<?>[])req.getOptionMap().get(Packet.OPTION_PARAMTYPES);
				for(int i = 0; i < args.length; i++) {
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}
			else {
				for(int i = 0; i < args.length; i++) {
					paramType[i] = Packet.convertPrimitive(params[i+offset]);
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}

			Method method = null;
			try {
				// interfaceをインプリしていても適合だと判断されない。
				method = c.getMethod(methodName, paramType);
			}
			catch(NoSuchMethodException e) {
				Logger.debug("NoSuchMethodException, so finding other way", e);
				Method[] methods = c.getMethods();
				for(int i = 0; i < methods.length; i++) {
					if(methods[i].getName().equals(methodName) && methods[i].getParameterTypes().length == args.length ) {
						// メソッド名とパラメータ数が一致すれば適合と見なす
						method = methods[i];
						break;
					}
				}
			}

			// リスナ登録が必要な場合は登録する
			setListener(req, args);

			if(method == null) {
				throw new NoSuchMethodException("not found method, className["+className+"] methodName["+methodName+"]");
			}
			result = method.invoke(object, args);

// インスタンスをクライアントに送るとクライアントでサーバのクラスが必要になるし、
// そのオブジェクトが生成されてしまうのも良くない。
//			// シングルトンの場合はインスタンスを返却しない。（インスタンスを生成できないと送れない。）
//			if(req.getInstanceNumber() > 0) {
//				req.setInstance(object);
//			}
			req.setResult(result);
		}

		/**
		 * RciListenerInvokerから非同期通知をクライアントに転送するためにこのメソッドを使う。
		 *
		 * @param listenerKey クライアントでイベントリスナーを起動するための識別番号
		 * @param objects メソッド名とパラメータ
		 * @throws Exception 何らかの例外
		 */
		protected void invokeClientListener(int listenerKey, Object ... objects ) throws Exception{
			// クライアント側でメソッドを実行する
			Packet req = new Packet();
			req.setParams(objects);
			req.setInstanceNumber(listenerMap.get(listenerKey));
			req.getOptionMap().put(Packet.OPTION_CALLBACK, true);

			req.enSerializeParam();

			// serializeできていれば、それを使ってインスタンスを作るので、XMLEncoderの対象外にする。
			avoidXMLEncode(req);

			String sendData = Packet.getEncodeXML(req);
			send(sendData);
		}

		/**
		 * serializeできていれば、それを使ってインスタンスを作るので、XMLEncoderの対象外にする。
		 * @param req パケット
		 */
		protected void avoidXMLEncode(Packet req) {
			Object[] objects = req.getParams();
			Object[] serializedParams = req.getSerializedParams();
			for(int i = 0; i < serializedParams.length; i++) {
				if(serializedParams[i] != null) {
					objects[i] = null;
				}
			}
		}

		/**
		 * サーバ側でイベントをクライアントに転送するリスナーを登録する。
		 * @param req パケット
		 * @param args リスナー登録情報
		 */
		protected void setListener(Packet req, Object[] args) {
			Object obj = req.getOptionMap().get(Packet.OPTION_LISTENER_NUMBER);
			if(obj == null) {
				// registListenerではない
				return ;
			}

			int lNumber = (int) obj;
			for(int i = 0; i < args.length; i++) {
				if(args[i] instanceof EventListener) {
					listenerMap.put(args[i].hashCode(), lNumber);
					RciListenerInvoker.gi().getListenerInvoker().put(args[i].hashCode(), this);
					req.getOptionMap().put(Packet.OPTION_LISTENER_HASHCODE, args[i].hashCode());
				}
			}
		}

		/**
		 * サーバ側でイベントをクライアントに転送するリスナーを削除する。
		 * @param req パケット
		 */
		protected void removeListener(Packet req) {
			Object obj = req.getOptionMap().get(Packet.OPTION_LISTENER_NUMBER);
			if(obj == null) {
				// registListenerではない
				return ;
			}

			obj = req.getOptionMap().get(Packet.OPTION_LISTENER_HASHCODE);
			if(obj != null) {
				RciListenerInvoker.gi().getListenerInvoker().remove((int)obj);
			}
		}

		/**
		 * RciListenerInvokerのイベント転送情報を削除する。
		 */
		protected void delListener() {
			Integer[] key = listenerMap.keySet().toArray(new Integer[0]);
			for(int i = 0; i < key.length; i++) {
				RciListenerInvoker.gi().getListenerInvoker().remove(key[i]);
				listenerMap.remove(key[i]);
			}
		}

		/**
		 * 鍵交換を行う
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
//		protected void invokeExchangeKey(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		protected void invokeExchangeKey(Packet req) throws Exception {
			Object[] params = req.getParams();
			Object result;
			String publicKey = params[1].toString();
			cryptFilter.pairCrypt.setPublicKey(publicKey);

			String[] encrypted;
			int aesCryptCount = Cm.gi().getConfig().getServerConfig().getAesCryptCount();
			encrypted = new String[aesCryptCount*2];
			for(int i = 0; i < aesCryptCount; i++) {
				AesCrypt ac = new AesCrypt();
				cryptFilter.addAesCrypt(ac);

				byte[] secret = cryptFilter.pairCrypt.encryptByPublicKey(ac.getSecretKey());
				byte[] iv = cryptFilter.pairCrypt.encryptByPublicKey(ac.getIvs());
				encrypted[i*2] = new String(Base64.getEncoder().encode(secret));
				encrypted[i*2+1] = new String(Base64.getEncoder().encode(iv));
			}

			result = encrypted;
			req.setResult(result);
			Logger.info("exchange key successfully from["+remoteAddress+"] at["+this.toString()+"]");
		}

		/**
		 * ユーザ認証を行う
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
//		protected void invokeCertificateUser(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		protected void invokeCertificateUser(Packet req) throws Exception {
			Object[] params = req.getParams();
			Object result;
			String username = params[1].toString();
			String password = params[2].toString();


			// ユーザ情報が存在するか照合する。
			UserConfig userConfig = Cm.gi().getConfig().getUserNameFilter().get(username);
			if(userConfig == null) {
				 throw new IllegalAccessException("unknown user["+username+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			LoginFailureUser lfu = accepter.failLogin.get(username);
			if(lfu != null) {
				if(!lfu.isEnableLogin()) {
					// 第三者が嫌がらせしたら正規ユーザがログインできなくなるが。
					throw new IllegalAccessException("too much failed login, now reject user["+username+"] from["+remoteAddress+"] at["+this.toString()+"]");
				}
			}

			// ユーザ情報からpasswordを取得し、hashRepet分hash化し、パラメータと照合する。
			HashCrypt hc = new HashCrypt();
	    	boolean isCertificate = hc.certificate(userConfig.getHashAlgorithmName(), userConfig.getPasswordAuthHashRoundCount(), password, userConfig.getPasswordRegistHashRoundCount(), userConfig.getPassword());
	    	if(!isCertificate) {
	    		if(lfu == null) {
	    			lfu = new LoginFailureUser(username);
	    			accepter.failLogin.put(username, lfu);
	    		}
				lfu.loginFailed();
				throw new IllegalAccessException("password confirm failure user["+username+"] from["+remoteAddress+"] at["+this.toString()+"]");
	    	}
	    	else {
	    		accepter.failLogin.remove(username);
	    	}

			// 公開鍵がサーバの設定ファイルに登録されている場合は照合する。
	    	//（この場合、クライアントは対応する秘密鍵を持っていたことになるので、認証はとても固い）
			String configPublicKey = userConfig.getPublicKey();
			if(configPublicKey != null && !"".equals(configPublicKey)) {
				if(!configPublicKey.equals(cryptFilter.pairCrypt.getPublicKey())) {
					 throw new IllegalAccessException("publickey confirm failure user["+username+"] from["+remoteAddress+"] at["+this.toString()+"]");
				}
			}
			loginUsername = username;
			Logger.info("successfully logged in user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			result = "Certificated on "+ this.toString();
			req.setResult(result);
		}

		/**
		 * パスワード情報を更新する。
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
//		protected void invokeUpdatePassword(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException, IOException {
		protected void invokeUpdatePassword(Packet req) throws Exception {
			Object[] params = req.getParams();
			Object result;
			String hashAlgorithmName = (String)params[1];
			String loginPassword = (String)params[2];
			String registPassword = (String)params[3];

			int loginCount = (Integer)params[4];
			int registCount = (Integer)params[5];

			String pubKey = (String)params[6];

			String remoteAddress = socket.getInetAddress().getHostAddress();

			// ユーザ情報が存在するか照合する。
			UserConfig userConfig = Cm.gi().getConfig().getUserNameFilter().get(loginUsername);
			if(userConfig == null) {
				 throw new IllegalAccessException("update password failure, unknown user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			// 新パスワードでログイン可能な場合のみ更新する
			//（registCount > loginCountの場合のみ登録可能にし、registPasswordはユーザから与えられないようにすべきか。）
			HashCrypt hc = new HashCrypt();
	    	boolean isCertificate = hc.certificate(hashAlgorithmName, loginCount, loginPassword, registCount, registPassword);
	    	if(!isCertificate) {
				throw new IllegalAccessException("new password confirm failure user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
	    	}

			userConfig.setPassword(registPassword);
			userConfig.setHashAlgorithmName(hashAlgorithmName);
			userConfig.setPasswordAuthHashRoundCount(loginCount);
			userConfig.setPasswordRegistHashRoundCount(registCount);
			if(!StringUtils.isEmpty(pubKey)) {
				// ネットワーク越しに更新できるようにすると万一乗っ取られたら乗っ取り犯にお墨付きを与えることにならないか。
				userConfig.setPublicKey(pubKey);
			}

			// userConfigの内容をファイルに保存する
			Cm.gi().backupUserFile(loginUsername);
			Cm.gi().saveUser(loginUsername);

			Logger.info("successfully update password information["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			result = "Update password information on "+ this.toString();
			req.setResult(result);
		}

		// new className param1 param2
		/**
		 * クラスのインスタンスを生成する。
		 * @param req パケット
		 * @return インスタンス番号
		 * @throws Exception 何らかの例外
		 */
//		int invokeNew(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		int invokeNew(Packet req) throws Exception {
			Object[] params;
			params = req.getParams();

			Object result;
			String className = params[1].toString();
			String packageName = ClassField.getPackageName(className);
			if(!isAccessiblePackage(packageName)) {
				throw new IllegalAccessException("not allow package["+packageName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}
			if(!isAccessibleClass(className)) {
				throw new IllegalAccessException("not allow className["+className+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			Class<?> c = Class.forName(className);

			int offset = 2;
			Object[] args = new Object[params.length -offset];
			Class<?>[] paramType = new Class[params.length -offset];
			if(req.getOptionMap().containsKey(Packet.OPTION_PARAMTYPES)) {
				paramType = (Class<?>[])req.getOptionMap().get(Packet.OPTION_PARAMTYPES);
				for(int i = 0; i < args.length; i++) {
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}
			else {
				for(int i = 0; i < args.length; i++) {
					paramType[i] = Packet.convertPrimitive(params[i+offset]);
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}

			Constructor<?> constructor = null;
			try {
//				constructor = c.getDeclaredConstructor(paramType);
				constructor = c.getConstructor(paramType);
			}
			catch(NoSuchMethodException e) {
				Logger.debug("NoSuchMethodException, so finding other way", e);
				Constructor<?>[] constructors = c.getConstructors();
				for(int i = 0; i < constructors.length; i++) {
					if(constructors[i].getParameterTypes().length == args.length ) {
						// メソッド名とパラメータ数が一致すれば適合と見なす
						constructor = constructors[i];
						break;
					}
				}
			}
			if(constructor == null) {
				throw new NoSuchMethodException("not found constructor, className["+className+"]");
			}
			result = constructor.newInstance(args);

			instanceMap.put(++iNumber, result);
			req.setInstanceNumber(iNumber);

			// クライアントにインスタンスを送るとクライアントでそのクラスが生成されてしまうので良くない。
//			req.setInstance(result);
//			req.setResult(result);
			return iNumber;

		}

		// className methodName param1 param2
		/**
		 * スタティックメソッドを実行する。
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
//		protected void invokeStatic(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		protected void invokeStatic(Packet req) throws Exception {
			Object[] params = req.getParams();
			Object result;
			String className = params[1].toString();

			String packageName = ClassField.getPackageName(className);
			if(!isAccessiblePackage(packageName)) {
				throw new IllegalAccessException("not allow package["+packageName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			String methodName = params[2].toString();
			if(!isAccessibleClass(className, methodName)) {
				throw new IllegalAccessException("not allow className["+className+"] methodName["+methodName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			Class<?> c = Class.forName(className);

			int offset = 3;
			Object[] args = new Object[params.length -offset];
			Class<?>[] paramType = new Class[params.length -offset];
			if(req.getOptionMap().containsKey(Packet.OPTION_PARAMTYPES)) {
				paramType = (Class<?>[])req.getOptionMap().get(Packet.OPTION_PARAMTYPES);
				for(int i = 0; i < args.length; i++) {
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}
			else {
				for(int i = 0; i < args.length; i++) {
					paramType[i] = Packet.convertPrimitive(params[i+offset]);
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}

			Method method = null;
			try {
				// interfaceをインプリしていても適合だと判断されない。
				method = c.getMethod(methodName, paramType);
			}
			catch(NoSuchMethodException e) {
				Logger.debug("NoSuchMethodException, so finding other way", e);
				Method[] methods = c.getMethods();
				for(int i = 0; i < methods.length; i++) {
					if(methods[i].getName().equals(methodName) && methods[i].getParameterTypes().length == args.length ) {
						// メソッド名とパラメータ数が一致すれば適合と見なす
						method = methods[i];
						break;
					}
				}
			}

			if(method == null) {
				throw new NoSuchMethodException("not found method, className["+className+"] methodName["+methodName+"]");
			}
			result = method.invoke(null, args);
			req.setResult(result);
		}

		// className methodName param1 param2
		/**
		 * シングルトンのインスタンスを取得する
		 * @param req パケット
		 * @throws Exception 何らかの例外
		 */
//		protected void invokeSingle(Packet req) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException {
		protected void invokeSingle(Packet req) throws Exception {
			Object[] params = req.getParams();
			Object result;
			String className = params[1].toString();

			String methodName = params[2].toString();

			String packageName = ClassField.getPackageName(className);
			if(!isAccessiblePackage(packageName)) {
				throw new IllegalAccessException("not allow package["+packageName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}
			if(!isAccessibleClass(className, methodName)) {
				throw new IllegalAccessException("not allow className["+className+"] methodName["+methodName+"] user["+loginUsername+"] from["+remoteAddress+"] at["+this.toString()+"]");
			}

			Class<?> c = Class.forName(className);

			int offset = 3;
			Object[] args = new Object[params.length -offset];
			Class<?>[] paramType = new Class[params.length -offset];
			if(req.getOptionMap().containsKey(Packet.OPTION_PARAMTYPES)) {
				paramType = (Class<?>[])req.getOptionMap().get(Packet.OPTION_PARAMTYPES);
				for(int i = 0; i < args.length; i++) {
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}
			else {
				for(int i = 0; i < args.length; i++) {
					paramType[i] = Packet.convertPrimitive(params[i+offset]);
					args[i] = params[i+offset];
					if(args[i] instanceof NullObject) {
						args[i] = null;
					}
				}
			}

			Method method = null;
			try {
				// interfaceをインプリしていても適合だと判断されない。
				method = c.getMethod(methodName, paramType);
			}
			catch(NoSuchMethodException e) {
				Logger.debug("NoSuchMethodException, so finding other way", e);
				Method[] methods = c.getMethods();
				for(int i = 0; i < methods.length; i++) {
					if(methods[i].getName().equals(methodName) && methods[i].getParameterTypes().length == args.length ) {
						// メソッド名とパラメータ数が一致すれば適合と見なす
						method = methods[i];
						break;
					}
				}
			}

			if(method == null) {
				throw new NoSuchMethodException("not found method, className["+className+"] methodName["+methodName+"]");
			}
			result = method.invoke(null, args);

			instanceMap.put(--singleNumber, result);
			req.setInstanceNumber(singleNumber);
			req.setResult(iNumber);
		}

		/**
		 * クラスのアクセスが許容されているか判定する。
		 * @param className クラスの完全修飾名
		 * @return アクセス許可の場合はtrue
		 */
		protected boolean isAccessibleClass(String className) {
			UserConfig uc = null;
			if(userConfig == null) {
				userConfig = getUserConfig(loginUsername);
			}
			uc = userConfig;

			if(uc == null) {
				// ユーザ定義情報が取得できないような非想定の場合
				return false;
			}

			if( uc.getClassNameFilter().isEmpty()) {
				// クラス定義が空の場合は全クラスを許可する
				// (全クラスを不許可にするなら、存在しないクラス名を上位の親グループを含めて１つだけ置けばよい。)
				return true;
			}

			HashMap<String, String> methodMap = uc.getClassNameFilter().get(className);
			if( methodMap == null) {
//				Logger.warn("not allow class["+ className +"] by user["+loginUsername+"] config");
				return false;
			}
			return true;
		}

		/**
		 * パッケージのアクセスが許容されているか判定する。
		 * @param packageName パッケージ名
		 * @return アクセス許可の場合はtrue
		 */
		protected boolean isAccessiblePackage(String packageName) {
			UserConfig uc = null;
			if(userConfig == null) {
				userConfig = getUserConfig(loginUsername);
			}
			uc = userConfig;

			if(uc == null) {
				// ユーザ定義情報が取得できないような非想定の場合
				return false;
			}

			if( uc.getPackageDenyFilter().isEmpty()) {
				// 不許可定義がない
				if( uc.getPackageAllowFilter().isEmpty()) {
					// 不許可(0)、許可(0)→全許可
					return true;
				}
				else {
					String allow = uc.getPackageAllowFilter().get(packageName);
					if(allow == null) {
						// 不許可(0)、許可(非0、ハズレ)→不許可
						return false;
					}
					else {
						// 不許可(0)、許可(非0、ヒット)→許可
						return true;
					}
				}
			}
			else {
				// 不許可定義がある
				String deny = uc.getPackageDenyFilter().get(packageName);
				if(deny == null) {
					if( uc.getPackageAllowFilter().isEmpty()) {
						// 不許可(非0、ハズレ)、許可(0)→許可
						return true;
					}
					else {
						String allow = uc.getPackageAllowFilter().get(packageName);
						if(allow == null) {
							// 不許可(非0、ハズレ)、許可(非0、ハズレ)→不許可
							return false;
						}
						else {
							// 不許可(非0、ハズレ)、許可(非0、ヒット)→許可
							return true;
						}
					}
				}
				else {
					// 不許可(非0、ヒット)→不許可（許可と不許可に同じkeyがあるなら、不許可を優先）
					return false;
				}

			}
		}

		/**
		 * メソッドのアクセスが許容されているか判定する。
		 * @param className クラスの完全修飾名
		 * @param methodName メソッド名
		 * @return アクセス許可の場合はtrue
		 */
		protected boolean isAccessibleClass(String className, String methodName) {
			UserConfig uc = null;
			if(userConfig == null) {
				userConfig = getUserConfig(loginUsername);
			}
			uc = userConfig;

			if(uc == null) {
				// ユーザ定義情報が取得できないような非想定の場合
				return false;
			}

			if( uc.getClassNameFilter().isEmpty()) {
				// クラス定義が空の場合は全クラスを許可する
				// (全クラスを不許可にするなら、存在しないクラス名を上位の親グループを含めて１つだけ置けばよい。)
				return true;
			}

			HashMap<String, String> methodMap = uc.getClassNameFilter().get(className);
			if( methodMap == null) {
//				Logger.warn("not allow class["+ className +"] by user["+loginUsername+"] config");
				return false;
			}

			if(methodMap.isEmpty()) {
				// メソッド定義が空の場合は全メソッドを許可する
				// (クラス内の全メソッドを不許可にするなら、存在しないメソッド名を１つ置けばよい。)
			}
			else {
				String value = methodMap.get(methodName);
				if(value == null) {
//					Logger.warn("not allow method["+methodName+"] of class["+ className +"] by user["+loginUsername+"] config");
					return false;
				}
			}
			return true;
		}

		/**
		 * ユーザの設定情報を取得する。
		 * @param username ユーザ名
		 * @return UserConfig
		 */
		protected UserConfig getUserConfig(String username)  {
			UserConfig uc = new UserConfig();
			UserConfig ucConfig = Cm.gi().getConfig().getUserNameFilter().get(username);
			if(ucConfig == null) {
//				Logger.warn("unknown user["+ username +"] config");
				return null;
			}
			try {
				ClassField.clone(ucConfig,  uc);
			} catch (Exception e) {
				Logger.error("failed config clone for user["+ username +"]", e);
				return null;
			}

			for(String group = uc.getGroup(); !"".equals(group); group = StringUtils.getSeparatorParent(group,  "/")) {
				GroupConfig gc = Cm.gi().getConfig().getGroupFilter().get(group);
				if(gc != null) {
					// classNameFilter
					String[] gcClassNames = gc.getClassNameFilter().keySet().toArray(new String[0]);
					for(int i = 0; i < gcClassNames.length; i++) {
						HashMap<String, String> ucHash = uc.getClassNameFilter().get(gcClassNames[i]);
						if(ucHash == null) {
							HashMap<String, String> gcHash = gc.getClassNameFilter().get(gcClassNames[i]);
							if(gcHash != null) {
								// 子に存在しないクラスで親に存在する場合は追加する。（子が親をオーバライド）
								uc.getClassNameFilter().put(gcClassNames[i], gcHash);
							}
						}
					}

					// PackageDeny
					String[] gcPackageDenys = gc.getPackageDenyFilter().keySet().toArray(new String[0]);
					for(int i = 0; i < gcPackageDenys.length; i++) {
						String ucText = uc.getPackageDenyFilter().get(gcPackageDenys[i]);
						if(ucText == null) {
							String gcText = gc.getPackageDenyFilter().get(gcPackageDenys[i]);
							if(gcText != null) {
								// 子に存在せず親に存在する場合は追加する。（子が親をオーバライド）
								uc.getPackageDenyFilter().put(gcPackageDenys[i], gcText);
							}
						}
					}

					// PackageAllow
					String[] gcPackageAllows = gc.getPackageAllowFilter().keySet().toArray(new String[0]);
					for(int i = 0; i < gcPackageAllows.length; i++) {
						String ucText = uc.getPackageAllowFilter().get(gcPackageAllows[i]);
						if(ucText == null) {
							String gcText = gc.getPackageAllowFilter().get(gcPackageAllows[i]);
							if(gcText != null) {
								// 子に存在せず親に存在する場合は追加する。（子が親をオーバライド）
								uc.getPackageAllowFilter().put(gcPackageAllows[i], gcText);
							}
						}
					}
				}
			}
			return uc;
		}
	}

	/**
	 * 同じユーザ名でwatchingTimeMillis時間内にlimitFailCount回数ログインに失敗した場合は
	 * rejectTimeMillis時間ログインできないようにする
	 */
	class LoginFailureUser {
//		public static long DefaultWatchingTimeMillis = 60*60*1000;
//		public static long DefaultRejectTimeMillis = 60*60*1000;
//		public static int DefaultLimitFailCount = 10;
		protected long DefaultWatchingTimeMillis = Cm.gi().getConfig().getServerConfig().getLoginWatchingTimeMillis();
		protected long DefaultRejectTimeMillis = Cm.gi().getConfig().getServerConfig().getLoginRejectTimeMillis();
		protected int DefaultLimitFailCount = Cm.gi().getConfig().getServerConfig().getLoginLimitFailCount();

		protected String username = "";

		// 最初にログインに失敗した時刻
		protected long firstFailTimeMillis = 0;

		// 監視継続時間（デフォルトは1時間）
		protected long watchingTimeMillis = DefaultWatchingTimeMillis;

		// ログイン試行停止時間（デフォルトは1時間）
		protected long rejectTimeMillis = DefaultRejectTimeMillis;

		// ログイン試行停止解除時刻
		protected long expireRejectTimeMillis = 0;

		// 監視継続時間内に失敗したログイン数
		protected int failCount = 0;

		// 監視継続時間内にログイン失敗を許容する数(デフォルトは10回)
		protected int limitFailCount = DefaultLimitFailCount;

		LoginFailureUser() {
		}

		LoginFailureUser(String username) {
			this.username = username;
		}

		protected void loginFailed() {

			if(failCount == 0) {
				firstFailTimeMillis = System.currentTimeMillis();
			}
			long currentTime = System.currentTimeMillis();
			if(firstFailTimeMillis + watchingTimeMillis > currentTime) {
				failCount++;
			}

			if(failCount == limitFailCount) {
				Logger.warn("login fail limit count user["+username+"]");
				expireRejectTimeMillis = currentTime + rejectTimeMillis;
			}
		}

		protected boolean isEnableLogin() {
			long currentTime = System.currentTimeMillis();
			if(currentTime < expireRejectTimeMillis) {
				return false;
			}
			else {
				return true;
			}
		}

		/**
		 * @return username
		 */
		protected String getUsername() {
			return username;
		}

		/**
		 * @param username セットする username
		 */
		protected void setUsername(String username) {
			this.username = username;
		}

		/**
		 * @return firstFailTimeMillis
		 */
		protected long getFirstFailTimeMillis() {
			return firstFailTimeMillis;
		}

		/**
		 * @param firstFailTimeMillis セットする firstFailTimeMillis
		 */
		protected void setFirstFailTimeMillis(long firstFailTimeMillis) {
			this.firstFailTimeMillis = firstFailTimeMillis;
		}

		/**
		 * @return watchingTimeMillis
		 */
		protected long getWatchingTimeMillis() {
			return watchingTimeMillis;
		}

		/**
		 * @param watchingTimeMillis セットする watchingTimeMillis
		 */
		protected void setWatchingTimeMillis(long watchingTimeMillis) {
			this.watchingTimeMillis = watchingTimeMillis;
		}

		/**
		 * @return rejectTimeMillis
		 */
		protected long getRejectTimeMillis() {
			return rejectTimeMillis;
		}

		/**
		 * @param rejectTimeMillis セットする rejectTimeMillis
		 */
		protected void setRejectTimeMillis(long rejectTimeMillis) {
			this.rejectTimeMillis = rejectTimeMillis;
		}

		/**
		 * @return expireRejectTimeMillis
		 */
		protected long getExpireRejectTimeMillis() {
			return expireRejectTimeMillis;
		}

		/**
		 * @param expireRejectTimeMillis セットする expireRejectTimeMillis
		 */
		protected void setExpireRejectTimeMillis(long expireRejectTimeMillis) {
			this.expireRejectTimeMillis = expireRejectTimeMillis;
		}

		/**
		 * @return failCount
		 */
		protected int getFailCount() {
			return failCount;
		}

		/**
		 * @param failCount セットする failCount
		 */
		protected void setFailCount(int failCount) {
			this.failCount = failCount;
		}

		/**
		 * @return limitFailCount
		 */
		protected int getLimitFailCount() {
			return limitFailCount;
		}

		/**
		 * @param limitFailCount セットする limitFailCount
		 */
		protected void setLimitFailCount(int limitFailCount) {
			this.limitFailCount = limitFailCount;
		}

	}

}

